/*
Winding machine fully automatic with stepper motor to move the wire while wound by a continous rotation servo.
 
 
 
 This example code is in the public domain.
 
 http://www.arduino.cc/en/Tutorial/LiquidCrystal
 */
/* interrupt routine for Rotary Encoders
 tested with Noble RE0124PVB 17.7FINB-24 http://www.nobleusa.com/pdf/xre.pdf - available at pollin.de
 and a few others, seems pretty universal
 
 The average rotary encoder has three pins, seen from front: A C B
 Clockwise rotation A(on)->B(on)->A(off)->B(off)
 CounterCW rotation B(on)->A(on)->B(off)->A(off)
 
 and may be a push switch with another two pins, pulled low at pin 8 in this case
 raf@synapps.de 20120107
 */
// include the LCD library code:
#include <LiquidCrystal.h>

// initialize the library with the numbers of the interface pins
LiquidCrystal lcd(15, 14, 7, 6, 5, 4); //RS, Enable, D4, D5, D6, D7, R/W to GND


// encoder pins
enum PinAssignments {
  encoderPinA = 2,   // rigth
  encoderPinB = 3,   // left
  ENCODERBUTTON = 8    // button
};
//A4988 pins
#define DIR 13
#define STEP 12
//L293D pins
#define HBRIDGEA 11
#define HBRIDGEB 10
#define PWMPIN 9

const int QTRwindingPin = A3;
const int QTRwireLengthPin = A2;

//define the reflectance thresholds. For my setup using the white servo arm with 2mm space in between those values are ok 
#define QTRon 750
#define QTRoff 850
#define QTRWireLengthOn 500 //white surface
#define QTRWireLengthOff 700 //black surface

//variables for the QTR sensors
int QTRvalue, QTRwireLengthValue = 0;
boolean passed, wireLengthPassed = 0;
int windingCounter, wireLengthCounter =0;


// interrupt service routine vars
boolean A_set = false;              
boolean B_set = false;

//menu definitions
#define MOTOR_ON_MENU 0
#define PWM_MENU 1
#define COIL_LENGTH_MENU 3
#define WINDINGS_MENU 2
#define WIRE_MENU 4
//global control variables 
int menuCounter = 2; //start with something to adjust
boolean reversed = false;
boolean motoron = false;
boolean doneWinding = false;

//debounce variables for the encoder button
int ENCODERBUTTONstate;
int lastENCODERBUTTONstate = HIGH;
long lastDebounceTimeENCODERBUTTON = 0;
long debounceDelay = 50;

//machine parameters
float coilLength = 3.0;
float wireDiameter = 0.15;
float stepsPerMM = 200.0; //the M6 rod on the wire guide linear motion system has 1 mm pitch and we're using half step mode
float pulleyCircumference = 113.097; //the wire length measuring pulley has a radius of 18mm (from center to the deepest point in the grove), so the perimeter is 18.0 * pi * 2
float lengthWireWound =0.0;
float lengthWireAvailable =0.0;


volatile unsigned int encoderPWMPos = 70;  // a variable that persists throughout interrupt routine calls for the PWM
volatile unsigned int encoderWindingsPos = 10;
volatile unsigned int encoderLengthPos = 3;
volatile unsigned int encoderWirePos = 0;
unsigned int lastReportedPWMPos, lastReportedWindingsPos,  lastReportedLengthPos, lastReportedWirePos = 1;   // change management

static boolean rotating=false;      // debounce management
int remainingWindings=encoderWindingsPos;
long lastWindingTime = 0;
boolean layerToggle = true;
int currentPos=0;


void setup() {

  pinMode(encoderPinA, INPUT); 
  pinMode(encoderPinB, INPUT); 
  pinMode(ENCODERBUTTON, INPUT);
  // turn on pullup resistors
  digitalWrite(encoderPinA, HIGH);
  digitalWrite(encoderPinB, HIGH);
  digitalWrite(ENCODERBUTTON, HIGH);

  //H bridge control pins 
  pinMode(HBRIDGEA,OUTPUT);
  pinMode(HBRIDGEB,OUTPUT);
  //Stepper control Pins
  pinMode(STEP,OUTPUT);
  pinMode(DIR,OUTPUT);
 digitalWrite(DIR, LOW);
  // encoder pin on interrupt 0 (pin 2)
  attachInterrupt(0, doEncoderA, CHANGE);
  // encoder pin on interrupt 1 (pin 3)
  attachInterrupt(1, doEncoderB, CHANGE);
  // set up the LCD's number of columns and rows: 
  lcd.begin(16, 2);
  displayMenu();
  //show arrow on default menu start position, which is "number of windings"
  lcd.setCursor(0, 0);
  lcd.print(">");

}

void loop() {
  int ENCODERBUTTONreading = digitalRead(ENCODERBUTTON);
  rotating = true;  // reset the debouncer

  if (ENCODERBUTTONreading != lastENCODERBUTTONstate) {
    // reset the debouncing timer
    lastDebounceTimeENCODERBUTTON = millis();
  }

  //manage the encoder button here to switch between value entrie modes and motor start...
  if ((millis() - lastDebounceTimeENCODERBUTTON) > debounceDelay) {
    // whatever the reading is at, it's been there for longer
    // than the debounce delay, so take it as the actual current state:

    // if the button state has changed:
    if (ENCODERBUTTONreading != ENCODERBUTTONstate) {
      ENCODERBUTTONstate = ENCODERBUTTONreading;

      // only toggle the LED if the new button state is HIGH
      if (ENCODERBUTTONstate == LOW) {
        menuCounter++;
        if(menuCounter > 4)
          menuCounter = 0;

        switch(menuCounter){
          //set winding motor speed
        case PWM_MENU : 
          displayMenu();
          lcd.setCursor(0, 1);
          lcd.print(">");

          break;
          //set number of windings
        case WINDINGS_MENU : 
          displayMenu();
          lcd.setCursor(0, 0);
          lcd.print(">");
          break;
          //start
        case MOTOR_ON_MENU : 
          displayMenu();
          lcd.setCursor(13, 1);
          lcd.print(">");
          break;
          //coil length
        case COIL_LENGTH_MENU : 
          displayMenu();
          lcd.setCursor(8, 0);
          lcd.print(">");
          break;
          //wire diameter
        case WIRE_MENU : 
          displayMenu();
          lcd.setCursor(4, 1);
          lcd.print(">");
          break;



        }
      }
    }
  }
  //refresh the values on the display

  switch(menuCounter){
    //set winding motor speed
  case PWM_MENU : 
    if (lastReportedPWMPos != encoderPWMPos) {
      if(encoderPWMPos > 255) encoderPWMPos = 0;
      else if(encoderPWMPos < 1) encoderPWMPos = 255;
      lastReportedPWMPos = encoderPWMPos;
    }
    displayPWM(encoderPWMPos);

    //PWM adjustment should also be accessible in motoron=true mode
    if(doneWinding == true) {
      motoron = false;
      encoderWindingsPos = 0;
      windingCounter = 0;
    }
    else motoron = true;
    break;

  case WINDINGS_MENU:
    if (lastReportedWindingsPos != encoderWindingsPos) {
      if(encoderWindingsPos > 65534) encoderWindingsPos = 0;
      lastReportedWindingsPos = encoderWindingsPos;
    }
    displayWindings(remainingWindings);
    motoron=false;
    break;

  case WIRE_MENU:
    if (lastReportedWirePos != encoderWirePos) {
      if(encoderWirePos > 65534) encoderWirePos = 0;
      lastReportedWirePos = encoderWirePos;
    }
    switch (encoderWirePos){
    case 0: 
      wireDiameter = 0.05;
      break;
    case 1: 
      wireDiameter = 0.1;
      break;
    case 2: 
      wireDiameter = 0.15;
      break;
    case 3: 
      wireDiameter = 0.2;
      break;
    case 4: 
      wireDiameter = 0.3;
      break;
    case 5: 
      wireDiameter = 0.4;
      break;
    case 6: 
      wireDiameter = 0.5;
      break;
    case 7: 
      wireDiameter = 0.6;
      break;
    case 8: 
      wireDiameter = 0.8;
      break;
    }
    if(encoderWirePos > 8) encoderWirePos = 0;
    else if(encoderWirePos < 0) encoderWirePos = 8;
    displayWire(wireDiameter);
    motoron=false;
    break;

  case COIL_LENGTH_MENU:
    if (lastReportedLengthPos != encoderLengthPos) {
      if(encoderLengthPos > 110) encoderLengthPos = 0;
      else if(encoderLengthPos == 0) encoderLengthPos = 110;
      lastReportedLengthPos = encoderLengthPos;
    }
    displayLength(encoderLengthPos);
    motoron=false;
    break;

  case MOTOR_ON_MENU:
    if(doneWinding == true) {
      motoron = false;
      encoderWindingsPos = 0;
      windingCounter = 0;
    }
    else motoron = true;
    break;

    //in all other cases turn off the winding motor
  default:
    motoron=false;

  }

  // read the QTR sensor attached to the winding motor
  QTRvalue = analogRead(QTRwindingPin); 

  if(QTRvalue < QTRon && passed == true){
    if(reversed==false){
      windingCounter ++; 
      currentPos++;
      //move slide one winding further
      for(int i= 0; i<int(float(stepsPerMM) * wireDiameter); i++){
        digitalWrite(STEP, HIGH);
        delay(1);
        digitalWrite(STEP, LOW);
        delay(1);
      }
      //lastWindingTime=millis();
      if (currentPos==int(float(encoderLengthPos)/wireDiameter)){ 
      digitalWrite(DIR, layerToggle);
      layerToggle = !layerToggle;
      currentPos =0;
      }
    }else windingCounter --;
      
    displayWindings(remainingWindings);
    //if(windingCounter < 1) windingCounter = 0;
    passed = false; 
  }

  if(QTRvalue > QTRoff && passed == false){   
    passed = true;
  }
  
    // read the QTR sensor attached to the wire length measuring pulley
  QTRwireLengthValue = analogRead(QTRwireLengthPin); 

  if(QTRwireLengthValue < QTRWireLengthOn && wireLengthPassed == true){
    if(reversed==false)
      wireLengthCounter ++; 
    else
      wireLengthCounter --;
      
    lengthWireWound = pulleyCircumference / 4.0 * (float) wireLengthCounter / 1000.0; //display used wire in [m] the 4 .0 is how often the counter state changes
    displayWireLengthWound(lengthWireWound);
    wireLengthPassed = false; 
  }

  if(QTRwireLengthValue > QTRWireLengthOn && wireLengthPassed == false){   
    wireLengthPassed = true;
  }

  //check if the entered number of windings wqas wound
  if(remainingWindings == 0) {
    doneWinding=true;
     digitalWrite(DIR, LOW);
    layerToggle=true;
  }
  else doneWinding = false;

  //check if motor is supposed to be on first
  if(motoron == true){
    //then check into which direction to turn
    if (reversed==true ){
      windClockwise();
    }
    else{
      windCounterClockwise();
    }
  }
  else{
    stopWinding();
  }

  analogWrite(PWMPIN,encoderPWMPos);

  remainingWindings=encoderWindingsPos-windingCounter;
  // save the reading.  Next time through the loop,
  // it'll be the lastButtonState:
  lastENCODERBUTTONstate = ENCODERBUTTONreading;
}//end loop***********************************************************************************************************************

//three functions for the H bridge IC L293D: one direction, other direction and stop 
void windClockwise(){
  digitalWrite(HBRIDGEA,LOW);
  digitalWrite(HBRIDGEB,HIGH);
}

void windCounterClockwise(){
  digitalWrite(HBRIDGEA,HIGH);
  digitalWrite(HBRIDGEB,LOW);
}

void stopWinding(){
  digitalWrite(HBRIDGEA,LOW);
  digitalWrite(HBRIDGEB,LOW);
}

void displayMenu(){
  lcd.clear();
  // set the cursor to column 0, line 1
  // (note: line 1 is the second row, since counting begins with 0):

  //set the winding motor speed, save space, say nothing
  //lcd.setCursor(1, 1);
  //lcd.print("PWM:");
  displayPWM(encoderPWMPos); 
  //set the desired number of windings
  lcd.setCursor(1, 0);
  lcd.print("n=");
  displayWindings(remainingWindings);
  //start winding tag  
  lcd.setCursor(14, 1);
  lcd.print("go");
  //display used wire
  lcd.setCursor(5, 1);
  lcd.print(lengthWireWound);
  //coil former length
  lcd.setCursor(9, 0);
  lcd.print("l=");
  displayLength(encoderLengthPos);
}

void displayPWM(int desiredPWM){
  lcd.setCursor(1,1);
  //display the value and if needed delete the digits from last screen update
  if(desiredPWM>99){
    lcd.print(desiredPWM); 
  }
  else if(desiredPWM>9){
    lcd.print(desiredPWM); 
    lcd.print(" "); 
  }
  else {
    lcd.print(desiredPWM); 
    lcd.print("  "); 
  }
}

void displayWindings(int remaining){

  //display Windings properly
  lcd.setCursor(3,0);
  //display the value and if needed delete the digits from last screen update
  if(remaining>9999){
    lcd.print(remaining); 
  }
  else if(remaining>999){
    lcd.print(remaining); 
    lcd.print(" "); 
  }
  else  if(remaining){
    lcd.print(remaining); 
    lcd.print("  "); 
  } 
  else  if(remaining){
    lcd.print(remaining); 
    lcd.print("   ");
  }
  else {
    lcd.print(remaining); 
    lcd.print("    ");
  }
}

void displayLength(int lengthMM){
  lcd.setCursor(11,0);
  //display the value and if needed delete the digits from last screen update
  if(lengthMM){
    lcd.print(lengthMM); 
    lcd.print("  "); 
  } 
  else  if(lengthMM){
    lcd.print(lengthMM); 
    lcd.print("   ");
  }
  else {
    lcd.print(lengthMM); 
    lcd.print("    ");
  }
}

void displayWire(float diameter){
  lcd.setCursor(5,1);
  if(encoderWirePos==0 || encoderWirePos == 2){
    lcd.print(diameter); 
    lcd.print(" ");
  } else  lcd.print(diameter); 
}

//while winding display used-up wire instead of wire diameter
void displayWireLengthWound(float wireWound){
  lcd.setCursor(5,1);
  lcd.print(wireWound);
 lcd.print("m"); 
}

// Interrupt on A changing state
void doEncoderA(){
  // debounce
  if ( rotating ) delay (1);  // wait a little until the bouncing is done

  // Test transition, did things really change? 
  if( digitalRead(encoderPinA) != A_set ) {  // debounce once more
    A_set = !A_set;

    // adjust counter + if A leads B
    if ( A_set && !B_set ) 
      //PWM
      if(menuCounter == PWM_MENU)  {
        encoderPWMPos -= 5;
      }
      else if(menuCounter == WINDINGS_MENU){
        //Windings
        encoderWindingsPos -= 5;
      }
      else if(menuCounter==COIL_LENGTH_MENU){
        encoderLengthPos -= 1;
      }
      else if(menuCounter==WIRE_MENU){
        encoderWirePos -= 1;
      }

    rotating = false;  // no more debouncing until loop() hits again
  }
}

// Interrupt on B changing state, same as A above
void doEncoderB(){
  if ( rotating ) delay (1);
  if( digitalRead(encoderPinB) != B_set ) {
    B_set = !B_set;
    //  adjust counter - 1 if B leads A
    if( B_set && !A_set )
      //PWM 
      if(menuCounter == PWM_MENU){
        encoderPWMPos += 5;
      }
      else if(menuCounter == WINDINGS_MENU){
        encoderWindingsPos += 5;
      }
      else if(menuCounter==COIL_LENGTH_MENU){
        encoderLengthPos += 1;
      }
      else if(menuCounter==WIRE_MENU){
        encoderWirePos += 1;
      }

    rotating = false;
  }
}












